;; Generated package description from setup.el  -*- no-byte-compile: t -*-
(define-package "setup" "1.5.0.0.20251027.195908" "Helpful Configuration Macro" '((emacs "26.1")) :commit "de9b86c6f29cf9888e911924ffc4ce47312320c3" :authors '(("Philip Kaludercic" . "philipk@posteo.net")) :maintainer '("Philip Kaludercic" . "philipk@posteo.net") :keywords '("lisp" "local") :url "https://codeberg.org/pkal/setup.el")
